/*
<:copyright-broadcom 
 
 Copyright (c) 2002 Broadcom Corporation 
 All Rights Reserved 
 No portions of this material may be reproduced in any form without the 
 written permission of: 
          Broadcom Corporation 
          16215 Alton Parkway 
          Irvine, California 92619 
 All information contained in this document is Broadcom Corporation 
 company private, proprietary, and trade secret. 
 
:>
*/
#ifndef _BCMENET_H_
#define _BCMENET_H_

#include <linux/skbuff.h>
#include <linux/if_ether.h>
#if defined(CONFIG_BCM96338)
#include <6338_map.h>
#endif
#if defined(CONFIG_BCM96348)
#include <6348_map.h>
#endif
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include "boardparms.h"
#include <bcmnet.h>

/*---------------------------------------------------------------------*/
/* specify number of BDs and buffers to use                            */
/*---------------------------------------------------------------------*/
/* In order for ATM shaping to work correctly,
 * the number of receive BDS/buffers = # tx queues * # buffers per tx queue
 * (80 ENET buffers = 8 tx queues * 10 buffers per tx queue)
 */
#define NR_RX_BDS               80
#define NR_TX_BDS               200
#define ENET_MAX_MTU_SIZE       1522    /* Body(1500) + EH_SIZE(14) + VLANTAG(4)+ FCS(4) */
#define DMA_MAX_BURST_LENGTH    16      /* in 32 bit words */
#define RX_ENET_SKB_RESERVE     16
#define SKB_ALIGNED_SIZE        ((sizeof(struct sk_buff) + 0x0f) & ~0x0f)
#define RX_BUF_LEN              ((ENET_MAX_MTU_SIZE + (DMA_MAX_BURST_LENGTH * 4)) & ~((DMA_MAX_BURST_LENGTH * 4)-1))
#define RX_BUF_SIZE             (SKB_DATA_ALIGN(RX_BUF_LEN + RX_ENET_SKB_RESERVE) + sizeof(struct skb_shared_info))

/* misc. configuration */
#define DMA_FC_THRESH_LO        5
#define DMA_FC_THRESH_HI        10

/* IEEE 802.3 Ethernet constant */
#define ETH_CRC_LEN             4
#define ETH_MULTICAST_BIT       0x01

#define CACHE_TO_NONCACHE(x)    KSEG1ADDR(x)
#define NONCACHE_TO_CACHE(x)    KSEG0ADDR(x)

#define ERROR(x)        printk x
#ifndef ASSERT
#define ASSERT(x)       if (x); else ERROR(("assert: "__FILE__" line %d\n", __LINE__)); 
#endif

#if defined(DUMP_TRACE)
#define TRACE(x)        printk x
#else
#define TRACE(x)
#endif

#define NUM_CHANS_PER_EMAC          2
#if defined (CONFIG_BCM96348)
#define MAX_EMACS                   2
#else
#define MAX_EMACS                   1
#endif
#define EMAC1_RX_CHAN               0
#define EMAC1_TX_CHAN               1
#define EMAC2_RX_CHAN               2
#define EMAC2_TX_CHAN               3

typedef struct ethsw_info_t {
    int cid;                            /* Current chip ID */
    int page;                           /* Current page */
    int type;                           /* Ethernet Switch type */
} ethsw_info_t;

typedef struct emac_pm_addr_t {
    BOOL                valid;          /* 1 indicates the corresponding address is valid */
    unsigned int        ref;            /* reference count */
    unsigned char       dAddr[ETH_ALEN];/* perfect match register's destination address */
    char                unused[2];      /* pad */
} emac_pm_addr_t;                    
#define MAX_PMADDR          4           /* # of perfect match address */
/*
 * device context
 */ 
typedef struct BcmEnet_devctrl {
    struct net_device *dev;             /* ptr to net_device */
    struct net_device *next_dev;        /* next device */
    int             unit;               /* device control index */
    struct timer_list timer;            /* used by Tx reclaim */
    int             linkstatus_polltimer;
    int             linkstatus_phyport;
    int             linkstatus_holder;

    struct net_device_stats stats;      /* statistics used by the kernel */
    IOCTL_MIB_INFO MibInfo;

    volatile EmacRegisters *emac;       /* EMAC register base address */

    volatile DmaRegs *dmaCtrl;          /* EMAC DMA register base address */

    /* transmit variables */
    volatile DmaChannelCfg *txDma;      /* location of transmit DMA register set */
    volatile DmaDesc *txBds;            /* Memory location of tx Dma BD ring */
    struct sk_buff *txSkb[NR_TX_BDS];   /* list of SKBs pending transition */
    int      txFreeBds;                 /* # of free transmit bds */
    int      txHeadIndex;
    int      txTailIndex;

    /* receive variables */
    volatile DmaChannelCfg *rxDma;      /* location of receive DMA register set */
    volatile DmaDesc *rxBds;            /* Memory location of rx bd ring */
    int      rxAssignedBds;             /* # of BDs given to HW */
    int      rxHeadIndex;
    int      rxTailIndex;
    unsigned char   *buf_pool[NR_RX_BDS]; /* rx buffer pool */

    uint16          chipId;             /* chip's id */
    uint16          chipRev;            /* step */
    int             rxIrq;              /* rx dma irq */
    emac_pm_addr_t  pmAddr[MAX_PMADDR]; /* perfect match address */
    struct tasklet_struct task;         /* tasklet */
    int             linkState;          /* link status */
    ethsw_info_t    ethSwitch;          /* external switch */
    ETHERNET_MAC_INFO EnetInfo;
    unsigned int    vid;
    struct sk_buff *freeSkbList;
    unsigned char skbs[(NR_RX_BDS * SKB_ALIGNED_SIZE) + 0x10];
} BcmEnet_devctrl;

// BD macros
#define IncTxBDptr(x, s) if (x == ((BcmEnet_devctrl *)s)->txLastbdPtr) \
                             x = ((BcmEnet_devctrl *)s)->txFirstbdPtr; \
                      else x++

#define IncRxBDptr(x, s) if (x == ((BcmEnet_devctrl *)s)->rxLastBdPtr) \
                             x = ((BcmEnet_devctrl *)s)->rxFirstBdPtr; \
                      else x++

#define CARDNAME    "BCM63xx_ENET"

#endif /* _BCMENET_H_ */

